<?php

namespace Drupal\cached_computed_field;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;

/**
 * Service for managing cached computed fields.
 */
class CachedComputedFieldManager implements CachedComputedFieldManagerInterface {

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The field type plugin manager.
   *
   * @var \Drupal\Core\Field\FieldTypePluginManagerInterface
   */
  protected $fieldTypePluginManager;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * The config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Constructs a CachedComputedFieldManager object.
   *
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *   The entity field manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Core\Field\FieldTypePluginManagerInterface $fieldTypePluginManager
   *   The field type plugin manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory.
   */
  public function __construct(EntityFieldManagerInterface $entityFieldManager, EntityTypeManagerInterface $entityTypeManager, FieldTypePluginManagerInterface $fieldTypePluginManager, TimeInterface $time, ConfigFactoryInterface $configFactory) {
    $this->entityFieldManager = $entityFieldManager;
    $this->entityTypeManager = $entityTypeManager;
    $this->fieldTypePluginManager = $fieldTypePluginManager;
    $this->time = $time;
    $this->configFactory = $configFactory;
  }

  /**
   * Returns an array containing metadata of expired field items.
   *
   * @return array
   *   An indexed array containing information about expired field items, sorted
   *   by expiration date, each value an associative array with the following
   *   keys:
   *   - entity_type: The entity type ID of the entity that contains the expired
   *     field.
   *   - entity_id: The ID of the entity that contains the expired field.
   *   - field_name: The name of the expired field.
   *   - expire: The UNIX time stamp indicating when this field has expired.
   */
  public function getExpiredItems() {
    $request_time = $this->time->getRequestTime();
    $items_to_process = $this->configFactory->get('cached_computed_field.settings')->get('item_limit');

    $items = [];

    foreach ($this->getFieldMap() as $entity_type_id => $fields) {
      $storage = $this->entityTypeManager->getStorage($entity_type_id);
      $definition = $this->entityTypeManager->getDefinition($entity_type_id);
      foreach ($fields as $field_name => $field_info) {
        // Stop adding items to the queue when the limit has been reached.
        if ($items_to_process === 0) {
          break 2;
        }
        $query = $storage->getQuery();

        // Disable access checks. Cron is usually run as an anonymous user, and
        // we also want to make sure any content that is not accessible by
        // anonymous users is updated.
        $query->accessCheck(FALSE);

        // Fetch all entities that have expired fields, as well as the ones that
        // have never been processed yet.
        $condition = $query->orConditionGroup()
          ->condition("$field_name.expire", $request_time, '<')
          ->notExists("$field_name.expire")
          ->notExists($field_name);
        $query->condition($condition);
        if ($definition->hasKey('bundle')) {
          $query->condition($definition->getKey('bundle'), $field_info['bundles'], 'IN');
        }

        // Don't ask for more than we can chew.
        if ($items_to_process !== -1) {
          $query->range(0, $items_to_process);
        }

        /** @var \Drupal\Core\Entity\ContentEntityInterface $entity */
        foreach ($storage->loadMultiple($query->execute()) as $entity) {
          // Get the expiration time from the first field value so we can order
          // by it.
          $field_value = $entity->get($field_name)->first();
          $expire = !empty($field_value) ? $field_value->getValue()['expire'] : 0;

          $items[] = [
            'entity_type' => $entity_type_id,
            'entity_id' => $entity->id(),
            'field_name' => $field_name,
            'expire' => $expire,
          ];

          $items_to_process--;
        }
      }
    }

    usort($items, function ($a, $b) {
      // Poor man's spaceship operator, still supporting PHP 5.6 :(
      if ($a['expire'] == $b['expire']) {
        return 0;
      }
      return $a['expire'] < $b['expire'] ? -1 : 1;
    });

    return $items;
  }

  /**
   * Returns a lightweight map of cached computed fields across bundles.
   *
   * @return array
   *   An array keyed by entity type. Each value is an array which keys are
   *   field names and value is an array with two entries:
   *   - type: The field type.
   *   - bundles: An associative array of the bundles in which the field
   *     appears, where the keys and values are both the bundle's machine name.
   */
  public function getFieldMap() {
    $map = [];
    foreach ($this->getFieldTypes() as $field_type) {
      $map = array_merge_recursive($this->entityFieldManager->getFieldMapByFieldType($field_type), $map);
    }
    return $map;
  }

  /**
   * Returns a list of field types provided by the Cached Computed Field module.
   *
   * @return array
   *   The list of field types.
   */
  public function getFieldTypes() {
    return array_keys(array_filter($this->fieldTypePluginManager->getDefinitions(), function ($definition) {
      return $definition['provider'] === 'cached_computed_field';
    }));
  }

}
