<?php

namespace Drupal\cached_computed_field\EventSubscriber;

use Drupal\cached_computed_field\Event\RefreshExpiredFieldEventInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;

/**
 * Base class for event subscribers that refresh expired cached computed fields.
 */
abstract class RefreshExpiredFieldSubscriberBase implements RefreshExpiredFieldSubscriberInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The system time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * Constructs a new RefreshExpiredFieldSubscriberBase object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The system time service.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, TimeInterface $time) {
    $this->entityTypeManager = $entityTypeManager;
    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [RefreshExpiredFieldEventInterface::EVENT_NAME => [['refreshExpiredField']]];
  }

  /**
   * {@inheritdoc}
   */
  public function getEntity(RefreshExpiredFieldEventInterface $event) {
    return $this->entityTypeManager->getStorage($event->getEntityTypeId())->load($event->getEntityId());
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldDefinition(RefreshExpiredFieldEventInterface $event) {
    return $this->getEntity($event)->getFieldDefinition($event->getFieldName());
  }

  /**
   * {@inheritdoc}
   */
  public function getExpiredFieldValue(RefreshExpiredFieldEventInterface $event) {
    return $this->getEntity($event)->get($event->getFieldName())->getValue();
  }

  /**
   * {@inheritdoc}
   */
  public function updateFieldValue(RefreshExpiredFieldEventInterface $event, $value) {
    $request_time = $this->time->getRequestTime();
    $cache_lifetime = $this->getField($event)->getSettings()['cache-max-age'];

    $entity = $this->getEntity($event);
    $entity->set($event->getFieldName(), [
      'value' => $value,
      'expire' => $request_time + $cache_lifetime,
    ]);
    $entity->save();
  }

  /**
   * {@inheritdoc}
   */
  public function fieldNeedsRefresh(RefreshExpiredFieldEventInterface $event) {
    $field = $this->getField($event);
    // If the field has no value yet it is new and needs to be refreshed.
    if ($field->isEmpty()) {
      return TRUE;
    }

    // If the expire time has passed it needs to be refreshed.
    return $field->getValue()->expire < $this->time->getRequestTime();
  }

  /**
   * {@inheritdoc}
   */
  public function getField(RefreshExpiredFieldEventInterface $event) {
    return $this->getEntity($event)->get($event->getFieldName());
  }

}
